@@ -29,7 +29,8 @@
 #include "pagemap-cache.h"
 #include "fault-injection.h"
 #include "prctl.h"
-#include <compel/compel.h>
+#include "compel/infect-util.h"
+#include "pidfd-store.h"
 
 #include "protobuf.h"
 #include "images/pagemap.pb-c.h"
@@ -64,7 +65,7 @@ int do_task_reset_dirty_track(int pid)
 		if (errno == EINVAL) /* No clear-soft-dirty in kernel */
 			ret = 1;
 		else {
-			pr_perror("Can't reset %d's dirty memory tracker (%d)", pid, errno);
+			pr_perror("Can't reset %d's dirty memory tracker", pid);
 			ret = -1;
 		}
 	} else {
@@ -79,9 +80,8 @@ int do_task_reset_dirty_track(int pid)
 unsigned long dump_pages_args_size(struct vm_area_list *vmas)
 {
 	/* In the worst case I need one iovec for each page */
-	return sizeof(struct parasite_dump_pages_args) +
-		vmas->nr * sizeof(struct parasite_vma_entry) +
-		(vmas->nr_priv_pages + 1) * sizeof(struct iovec);
+	return sizeof(struct parasite_dump_pages_args) + vmas->nr * sizeof(struct parasite_vma_entry) +
+	       (vmas->nr_priv_pages + 1) * sizeof(struct iovec);
 }
 
 static inline bool __page_is_zero(u64 pme)
@@ -164,7 +164,8 @@ static bool is_stack(struct pstree_item *item, unsigned long vaddr)
  * the memory contents is present in the pagent image set.
  */
 
-static int generate_iovs(struct pstree_item *item, struct vma_area *vma, struct page_pipe *pp, u64 *map, u64 *off, bool has_parent)
+static int generate_iovs(struct pstree_item *item, struct vma_area *vma, struct page_pipe *pp, u64 *map, u64 *off,
+			 bool has_parent)
 {
 	u64 *at = &map[PAGE_PFN(*off)];
 	unsigned long pfn, nr_to_scan;
@@ -220,13 +221,12 @@ static int generate_iovs(struct pstree_item *item, struct vma_area *vma, struct
 	cnt_add(CNT_PAGES_LAZY, pages[1]);
 	cnt_add(CNT_PAGES_WRITTEN, pages[2]);
 
-	pr_info("Pagemap generated: %lu pages (%lu lazy) %lu holes\n",
-		pages[2] + pages[1], pages[1], pages[0]);
+	pr_info("Pagemap generated: %lu pages (%lu lazy) %lu holes\n", pages[2] + pages[1], pages[1], pages[0]);
 	return ret;
 }
 
-static struct parasite_dump_pages_args *prep_dump_pages_args(struct parasite_ctl *ctl,
-		struct vm_area_list *vma_area_list, bool skip_non_trackable)
+static struct parasite_dump_pages_args *
+prep_dump_pages_args(struct parasite_ctl *ctl, struct vm_area_list *vma_area_list, bool skip_non_trackable)
 {
 	struct parasite_dump_pages_args *args;
 	struct parasite_vma_entry *p_vma;
@@ -260,8 +260,7 @@ static struct parasite_dump_pages_args *prep_dump_pages_args(struct parasite_ctl
 	return args;
 }
 
-static int drain_pages(struct page_pipe *pp, struct parasite_ctl *ctl,
-		      struct parasite_dump_pages_args *args)
+static int drain_pages(struct page_pipe *pp, struct parasite_ctl *ctl, struct parasite_dump_pages_args *args)
 {
 	struct page_pipe_buf *ppb;
 	int ret = 0;
@@ -272,8 +271,8 @@ static int drain_pages(struct page_pipe *pp, struct parasite_ctl *ctl,
 	list_for_each_entry(ppb, &pp->bufs, l) {
 		args->nr_segs = ppb->nr_segs;
 		args->nr_pages = ppb->pages_in;
-		pr_debug("PPB: %d pages %d segs %u pipe %d off\n",
-				args->nr_pages, args->nr_segs, ppb->pipe_size, args->off);
+		pr_debug("PPB: %d pages %d segs %u pipe %d off\n", args->nr_pages, args->nr_segs, ppb->pipe_size,
+			 args->off);
 
 		ret = compel_rpc_call(PARASITE_CMD_DUMPPAGES, ctl);
 		if (ret < 0)
@@ -307,17 +306,19 @@ static int xfer_pages(struct page_pipe *pp, struct page_xfer *xfer)
 	return ret;
 }
 
-static int detect_pid_reuse(struct pstree_item *item,
-			    struct proc_pid_stat* pps,
-			    InventoryEntry *parent_ie)
+static int detect_pid_reuse(struct pstree_item *item, struct proc_pid_stat *pps, InventoryEntry *parent_ie)
 {
 	unsigned long long dump_ticks;
 	struct proc_pid_stat pps_buf;
 	unsigned long long tps; /* ticks per second */
 	int ret;
 
+	/* Check pid reuse using pidfds */
+	if (pidfd_store_ready())
+		return pidfd_store_check_pid_reuse(item->pid->real);
+
 	if (!parent_ie) {
-		pr_err("Pid-reuse detection failed: no parent inventory, " \
+		pr_err("Pid-reuse detection failed: no parent inventory, "
 		       "check warnings in get_parent_inventory\n");
 		return -1;
 	}
@@ -335,32 +336,72 @@ static int detect_pid_reuse(struct pstree_item *item,
 			return -1;
 	}
 
-	dump_ticks = parent_ie->dump_uptime/(USEC_PER_SEC/tps);
+	dump_ticks = parent_ie->dump_uptime / (USEC_PER_SEC / tps);
 
 	if (pps->start_time >= dump_ticks) {
 		/* Print "*" if unsure */
-		pr_warn("Pid reuse%s detected for pid %d\n",
-			pps->start_time == dump_ticks ? "*" : "",
-			item->pid->real);
+		pr_warn("Pid reuse%s detected for pid %d\n", pps->start_time == dump_ticks ? "*" : "", item->pid->real);
 		return 1;
 	}
 	return 0;
 }
 
-static int generate_vma_iovs(struct pstree_item *item, struct vma_area *vma,
-			     struct page_pipe *pp, struct page_xfer *xfer,
-			     struct parasite_dump_pages_args *args,
-			     struct parasite_ctl *ctl, pmc_t *pmc,
-			     bool has_parent, bool pre_dump)
+static int generate_vma_iovs(struct pstree_item *item, struct vma_area *vma, struct page_pipe *pp,
+			     struct page_xfer *xfer, struct parasite_dump_pages_args *args, struct parasite_ctl *ctl,
+			     pmc_t *pmc, bool has_parent, bool pre_dump, int parent_predump_mode)
 {
 	u64 off = 0;
 	u64 *map;
 	int ret;
 
-	if (!vma_area_is_private(vma, kdat.task_size) &&
-				!vma_area_is(vma, VMA_ANON_SHARED))
+	if (!vma_area_is_private(vma, kdat.task_size) && !vma_area_is(vma, VMA_ANON_SHARED))
 		return 0;
 
+	/*
+	 * To facilitate any combination of pre-dump modes to run after
+	 * one another, we need to take extra care as discussed below.
+	 *
+	 * The SPLICE mode pre-dump, processes all type of memory regions,
+	 * whereas READ mode pre-dump skips processing those memory regions
+	 * which lacks PROT_READ flag.
+	 *
+	 * Now on mixing pre-dump modes:
+	 * 	If SPLICE mode follows SPLICE mode	: no issue
+	 *		-> everything dumped both the times
+	 *
+	 * 	If READ mode follows READ mode		: no issue
+	 *		-> non-PROT_READ skipped both the time
+	 *
+	 * 	If READ mode follows SPLICE mode   	: no issue
+	 *		-> everything dumped at first,
+	 *		   the non-PROT_READ skipped later
+	 *
+	 * 	If SPLICE mode follows READ mode   	: Need special care
+	 *
+	 * If READ pre-dump happens first, then it has skipped processing
+	 * non-PROT_READ regions. Following SPLICE pre-dump expects pagemap
+	 * entries for all mappings in parent pagemap, but last READ mode
+	 * pre-dump cycle has skipped processing & pagemap generation for
+	 * non-PROT_READ regions. So SPLICE mode throws error of missing
+	 * pagemap entry for encountered non-PROT_READ mapping.
+	 *
+	 * To resolve this, the pre-dump-mode is stored in current pre-dump's
+	 * inventoy file. This pre-dump mode is read back from this file
+	 * (present in parent pre-dump dir) as parent-pre-dump-mode during
+	 * next pre-dump.
+	 *
+	 * If parent-pre-dump-mode and next-pre-dump-mode are in READ-mode ->
+	 * SPLICE-mode order, then SPLICE mode doesn't expect mappings for
+	 * non-PROT_READ regions in parent-image and marks "has_parent=false".
+	 */
+
+	if (!(vma->e->prot & PROT_READ)) {
+		if (opts.pre_dump_mode == PRE_DUMP_READ && pre_dump)
+			return 0;
+		if ((parent_predump_mode == PRE_DUMP_READ && opts.pre_dump_mode == PRE_DUMP_SPLICE) || !pre_dump)
+			has_parent = false;
+	}
+
 	if (vma_entry_is(vma->e, VMA_AREA_AIORING)) {
 		if (pre_dump)
 			return 0;
@@ -375,7 +416,7 @@ static int generate_vma_iovs(struct pstree_item *item, struct vma_area *vma,
 		return add_shmem_area(item->pid->real, vma->e, map);
 
 again:
-	ret = generate_iovs(item,vma, pp, map, &off, has_parent);
+	ret = generate_iovs(item, vma, pp, map, &off, has_parent);
 	if (ret == -EAGAIN) {
 		BUG_ON(!(pp->flags & PP_CHUNK_MODE));
 
@@ -391,11 +432,9 @@ again:
 	return ret;
 }
 
-static int __parasite_dump_pages_seized(struct pstree_item *item,
-		struct parasite_dump_pages_args *args,
-		struct vm_area_list *vma_area_list,
-		struct mem_dump_ctl *mdc,
-		struct parasite_ctl *ctl)
+static int __parasite_dump_pages_seized(struct pstree_item *item, struct parasite_dump_pages_args *args,
+					struct vm_area_list *vma_area_list, struct mem_dump_ctl *mdc,
+					struct parasite_ctl *ctl)
 {
 	pmc_t pmc = PMC_INIT;
 	struct page_pipe *pp;
@@ -406,6 +445,7 @@ static int __parasite_dump_pages_seized(struct pstree_item *item,
 	unsigned long pmc_size;
 	int possible_pid_reuse = 0;
 	bool has_parent;
+	int parent_predump_mode = -1;
 
 	pr_info("\n");
 	pr_info("Dumping pages (type: %d pid: %d)\n", CR_FD_PAGES, item->pid->real);
@@ -413,17 +453,14 @@ static int __parasite_dump_pages_seized(struct pstree_item *item,
 
 	timing_start(TIME_MEMDUMP);
 
-	pr_debug("   Private vmas %lu/%lu pages\n",
-		 vma_area_list->nr_priv_pages_longest, vma_area_list->nr_priv_pages);
+	pr_debug("   Private vmas %lu/%lu pages\n", vma_area_list->nr_priv_pages_longest, vma_area_list->nr_priv_pages);
 
 	/*
 	 * Step 0 -- prepare
 	 */
 
-	pmc_size = max(vma_area_list->nr_priv_pages_longest,
-		       vma_area_list->nr_shared_pages_longest);
-	if (pmc_init(&pmc, item->pid->real, &vma_area_list->h,
-			 pmc_size * PAGE_SIZE))
+	pmc_size = max(vma_area_list->nr_priv_pages_longest, vma_area_list->nr_shared_pages_longest);
+	if (pmc_init(&pmc, item->pid->real, &vma_area_list->h, pmc_size * PAGE_SIZE))
 		return -1;
 
 	if (!(mdc->pre_dump || mdc->lazy))
@@ -433,9 +470,7 @@ static int __parasite_dump_pages_seized(struct pstree_item *item,
 		 * use, i.e. on non-lazy non-predump.
 		 */
 		cpp_flags |= PP_CHUNK_MODE;
-	pp = create_page_pipe(vma_area_list->nr_priv_pages,
-					    mdc->lazy ? NULL : pargs_iovs(args),
-					    cpp_flags);
+	pp = create_page_pipe(vma_area_list->nr_priv_pages, mdc->lazy ? NULL : pargs_iovs(args), cpp_flags);
 	if (!pp)
 		goto out;
 
@@ -460,29 +495,40 @@ static int __parasite_dump_pages_seized(struct pstree_item *item,
 	}
 
 	if (xfer.parent) {
-		possible_pid_reuse = detect_pid_reuse(item, mdc->stat,
-						      mdc->parent_ie);
+		possible_pid_reuse = detect_pid_reuse(item, mdc->stat, mdc->parent_ie);
 		if (possible_pid_reuse == -1)
 			goto out_xfer;
 	}
 
-
 	/*
 	 * Step 1 -- generate the pagemap
 	 */
 	args->off = 0;
 	has_parent = !!xfer.parent && !possible_pid_reuse;
+	if (mdc->parent_ie)
+		parent_predump_mode = mdc->parent_ie->pre_dump_mode;
+
 	list_for_each_entry(vma_area, &vma_area_list->h, list) {
-		ret = generate_vma_iovs(item, vma_area, pp, &xfer, args, ctl,
-					&pmc, has_parent, mdc->pre_dump);
+		ret = generate_vma_iovs(item, vma_area, pp, &xfer, args, ctl, &pmc, has_parent, mdc->pre_dump,
+					parent_predump_mode);
 		if (ret < 0)
 			goto out_xfer;
 	}
 
 	if (mdc->lazy)
-		memcpy(pargs_iovs(args), pp->iovs,
-		       sizeof(struct iovec) * pp->nr_iovs);
-	ret = drain_pages(pp, ctl, args);
+		memcpy(pargs_iovs(args), pp->iovs, sizeof(struct iovec) * pp->nr_iovs);
+
+	/*
+	 * Faking drain_pages for pre-dump here. Actual drain_pages for pre-dump
+	 * will happen after task unfreezing in cr_pre_dump_finish(). This is
+	 * actual optimization which reduces time for which process was frozen
+	 * during pre-dump.
+	 */
+	if (mdc->pre_dump && opts.pre_dump_mode == PRE_DUMP_READ)
+		ret = 0;
+	else
+		ret = drain_pages(pp, ctl, args);
+
 	if (!ret && !mdc->pre_dump)
 		ret = xfer_pages(pp, &xfer);
 	if (ret)
@@ -512,10 +558,8 @@ out:
 	return exit_code;
 }
 
-int parasite_dump_pages_seized(struct pstree_item *item,
-		struct vm_area_list *vma_area_list,
-		struct mem_dump_ctl *mdc,
-		struct parasite_ctl *ctl)
+int parasite_dump_pages_seized(struct pstree_item *item, struct vm_area_list *vma_area_list, struct mem_dump_ctl *mdc,
+			       struct parasite_ctl *ctl)
 {
 	int ret;
 	struct parasite_dump_pages_args *pargs;
@@ -528,13 +572,47 @@ int parasite_dump_pages_seized(struct pstree_item *item,
 	 * able to read the memory contents.
 	 *
 	 * Afterwards -- reprotect memory back.
+	 *
+	 * This step is required for "splice" mode pre-dump and dump.
+	 * Skip this step for "read" mode pre-dump.
+	 * "read" mode pre-dump delegates processing of non-PROT_READ
+	 * regions to dump stage. Adding PROT_READ works fine for
+	 * static processing (target process frozen during pre-dump)
+	 * and fails for dynamic as explained below.
+	 *
+	 * Consider following sequence of instances to reason, why
+	 * not to add PROT_READ in "read" mode pre-dump ?
+	 *
+	 *	CRIU- "read" pre-dump		    Target Process
+	 *
+	 *					1. Creates mapping M
+	 *					   without PROT_READ
+	 * 2. CRIU freezes target
+	 *    process
+	 * 3. Collect the mappings
+	 * 4. Add PROT_READ to M
+	 *    (non-PROT_READ region)
+	 * 5. CRIU unfreezes target
+	 *    process
+	 *					6. Add flag PROT_READ
+	 *					   to mapping M
+	 *					7. Revoke flag PROT_READ
+	 *					   from mapping M
+	 * 8. process_vm_readv tries
+	 *    to copy mapping M
+	 *    (believing M have
+	 *     PROT_READ flag)
+	 * 9. syscall fails to copy
+	 *    data from M
 	 */
 
-	pargs->add_prot = PROT_READ;
-	ret = compel_rpc_call_sync(PARASITE_CMD_MPROTECT_VMAS, ctl);
-	if (ret) {
-		pr_err("Can't dump unprotect vmas with parasite\n");
-		return ret;
+	if (!mdc->pre_dump || opts.pre_dump_mode == PRE_DUMP_SPLICE) {
+		pargs->add_prot = PROT_READ;
+		ret = compel_rpc_call_sync(PARASITE_CMD_MPROTECT_VMAS, ctl);
+		if (ret) {
+			pr_err("Can't dump unprotect vmas with parasite\n");
+			return ret;
+		}
 	}
 
 	if (fault_injected(FI_DUMP_PAGES)) {
@@ -549,10 +627,12 @@ int parasite_dump_pages_seized(struct pstree_item *item,
 		return ret;
 	}
 
-	pargs->add_prot = 0;
-	if (compel_rpc_call_sync(PARASITE_CMD_MPROTECT_VMAS, ctl)) {
-		pr_err("Can't rollback unprotected vmas with parasite\n");
-		ret = -1;
+	if (!mdc->pre_dump || opts.pre_dump_mode == PRE_DUMP_SPLICE) {
+		pargs->add_prot = 0;
+		if (compel_rpc_call_sync(PARASITE_CMD_MPROTECT_VMAS, ctl)) {
+			pr_err("Can't rollback unprotected vmas with parasite\n");
+			ret = -1;
+		}
 	}
 
 	return ret;
@@ -588,7 +668,6 @@ int prepare_mm_pid(struct pstree_item *i)
 			return -1;
 	}
 
-
 	while (vn < ri->mm->n_vmas || img != NULL) {
 		struct vma_area *vma;
 
@@ -617,12 +696,11 @@ int prepare_mm_pid(struct pstree_item *i)
 				ri->vmas.rst_priv_size += PAGE_SIZE;
 		}
 
-		pr_info("vma 0x%"PRIx64" 0x%"PRIx64"\n", vma->e->start, vma->e->end);
+		pr_info("vma 0x%" PRIx64 " 0x%" PRIx64 "\n", vma->e->start, vma->e->end);
 
 		if (vma_area_is(vma, VMA_ANON_SHARED))
 			ret = collect_shmem(pid, vma);
-		else if (vma_area_is(vma, VMA_FILE_PRIVATE) ||
-				vma_area_is(vma, VMA_FILE_SHARED))
+		else if (vma_area_is(vma, VMA_FILE_PRIVATE) || vma_area_is(vma, VMA_FILE_SHARED))
 			ret = collect_filemap(vma);
 		else if (vma_area_is(vma, VMA_AREA_SOCKET))
 			ret = collect_socket_map(vma);
@@ -662,7 +740,7 @@ static inline bool check_cow_vmas(struct vma_area *vma, struct vma_area *pvma)
 	if (!(vma->e->flags & MAP_ANONYMOUS) && vma->e->shmid != pvma->e->shmid)
 		return false;
 
-	pr_debug("Found two COW VMAs @0x%"PRIx64"-0x%"PRIx64"\n", vma->e->start, pvma->e->end);
+	pr_debug("Found two COW VMAs @0x%" PRIx64 "-0x%" PRIx64 "\n", vma->e->start, pvma->e->end);
 	return true;
 }
 
@@ -782,10 +860,8 @@ static int premap_private_vma(struct pstree_item *t, struct vma_area *vma, void
 		 * bits there. Ideally we'd check for the whole COW-chain
 		 * having any data in.
 		 */
-		addr = mmap(*tgt_addr, size,
-				vma->e->prot | PROT_WRITE,
-				vma->e->flags | MAP_FIXED | flag,
-				vma->e->fd, vma->e->pgoff);
+		addr = mmap(*tgt_addr, size, vma->e->prot | PROT_WRITE, vma->e->flags | MAP_FIXED | flag, vma->e->fd,
+			    vma->e->pgoff);
 
 		if (addr == MAP_FAILED) {
 			pr_perror("Unable to map ANON_VMA");
@@ -805,8 +881,7 @@ static int premap_private_vma(struct pstree_item *t, struct vma_area *vma, void
 		if (vma_has_guard_gap_hidden(vma))
 			paddr -= PAGE_SIZE;
 
-		addr = mremap(paddr, size, size,
-				MREMAP_FIXED | MREMAP_MAYMOVE, *tgt_addr);
+		addr = mremap(paddr, size, size, MREMAP_FIXED | MREMAP_MAYMOVE, *tgt_addr);
 		if (addr != *tgt_addr) {
 			pr_perror("Unable to remap a private vma");
 			return -1;
@@ -814,9 +889,9 @@ static int premap_private_vma(struct pstree_item *t, struct vma_area *vma, void
 	}
 
 	vma->e->status |= VMA_PREMMAPED;
-	vma->premmaped_addr = (unsigned long) addr;
-	pr_debug("\tpremap %#016"PRIx64"-%#016"PRIx64" -> %016lx\n",
-		vma->e->start, vma->e->end, (unsigned long)addr);
+	vma->premmaped_addr = (unsigned long)addr;
+	pr_debug("\tpremap %#016" PRIx64 "-%#016" PRIx64 " -> %016lx\n", vma->e->start, vma->e->end,
+		 (unsigned long)addr);
 
 	if (vma_has_guard_gap_hidden(vma)) { /* Skip guard page */
 		vma->e->start += PAGE_SIZE;
@@ -847,8 +922,7 @@ static inline bool vma_force_premap(struct vma_area *vma, struct list_head *head
 
 			prev = list_entry(vma->list.prev, struct vma_area, list);
 			if (prev->e->end == vma->e->start) {
-				pr_debug("Force premmap for 0x%"PRIx64":0x%"PRIx64"\n",
-						vma->e->start, vma->e->end);
+				pr_debug("Force premmap for 0x%" PRIx64 ":0x%" PRIx64 "\n", vma->e->start, vma->e->end);
 				return true;
 			}
 		}
@@ -865,15 +939,15 @@ static int task_size_check(pid_t pid, VmaEntry *entry)
 #ifdef __s390x__
 	if (entry->end <= kdat.task_size)
 		return 0;
-	pr_err("Can't restore high memory region %lx-%lx because kernel does only support vmas up to %lx\n", entry->start, entry->end, kdat.task_size);
+	pr_err("Can't restore high memory region %lx-%lx because kernel does only support vmas up to %lx\n",
+	       entry->start, entry->end, kdat.task_size);
 	return -1;
 #else
 	return 0;
 #endif
 }
 
-static int premap_priv_vmas(struct pstree_item *t, struct vm_area_list *vmas,
-		void **at, struct page_read *pr)
+static int premap_priv_vmas(struct pstree_item *t, struct vm_area_list *vmas, void **at, struct page_read *pr)
 {
 	struct vma_area *vma;
 	unsigned long pstart = 0;
@@ -995,13 +1069,11 @@ static int restore_priv_vma_content(struct pstree_item *t, struct page_read *pr)
 			}
 
 			if (!vma_area_is(vma, VMA_PREMMAPED)) {
-				unsigned long len = min_t(unsigned long,
-						(nr_pages - i) * PAGE_SIZE,
-						vma->e->end - va);
+				unsigned long len = min_t(unsigned long, (nr_pages - i) * PAGE_SIZE, vma->e->end - va);
 
 				if (vma->e->status & VMA_NO_PROT_WRITE) {
-					pr_debug("VMA 0x%"PRIx64":0x%"PRIx64" RO %#lx:%lu IO\n",
-							vma->e->start, vma->e->end, va, nr_pages);
+					pr_debug("VMA 0x%" PRIx64 ":0x%" PRIx64 " RO %#lx:%lu IO\n", vma->e->start,
+						 vma->e->end, va, nr_pages);
 					BUG();
 				}
 
@@ -1023,8 +1095,7 @@ static int restore_priv_vma_content(struct pstree_item *t, struct page_read *pr)
 			 */
 
 			off = (va - vma->e->start) / PAGE_SIZE;
-			p = decode_pointer((off) * PAGE_SIZE +
-					vma->premmaped_addr);
+			p = decode_pointer((off)*PAGE_SIZE + vma->premmaped_addr);
 
 			set_bit(off, vma->page_bitmap);
 			if (vma_inherited(vma)) {
@@ -1068,7 +1139,6 @@ static int restore_priv_vma_content(struct pstree_item *t, struct page_read *pr)
 
 				bitmap_set(vma->page_bitmap, off + 1, nr - 1);
 			}
-
 		}
 	}
 
@@ -1093,11 +1163,10 @@ err_read:
 			/* Find all pages, which are not shared with this child */
 			i = find_next_bit(vma->pvma->page_bitmap, size, i);
 
-			if ( i >= size)
+			if (i >= size)
 				break;
 
-			ret = madvise(addr + PAGE_SIZE * i,
-						PAGE_SIZE, MADV_DONTNEED);
+			ret = madvise(addr + PAGE_SIZE * i, PAGE_SIZE, MADV_DONTNEED);
 			if (ret < 0) {
 				pr_perror("madvise failed");
 				return -1;
@@ -1119,14 +1188,13 @@ err_read:
 	return 0;
 
 err_addr:
-	pr_err("Page entry address %lx outside of VMA %lx-%lx\n",
-	       va, (long)vma->e->start, (long)vma->e->end);
+	pr_err("Page entry address %lx outside of VMA %lx-%lx\n", va, (long)vma->e->start, (long)vma->e->end);
 	return -1;
 }
 
 static int maybe_disable_thp(struct pstree_item *t, struct page_read *pr)
 {
-	struct _MmEntry *mm = rsti(t)->mm;
+	MmEntry *mm = rsti(t)->mm;
 
 	/*
 	 * There is no need to disable it if the page read doesn't
@@ -1205,8 +1273,7 @@ int prepare_mappings(struct pstree_item *t)
 	if (old_premmapped_addr) {
 		ret = munmap(old_premmapped_addr, old_premmapped_len);
 		if (ret < 0)
-			pr_perror("Unable to unmap %p(%lx)",
-					old_premmapped_addr, old_premmapped_len);
+			pr_perror("Unable to unmap %p(%lx)", old_premmapped_addr, old_premmapped_len);
 	}
 
 	/*
@@ -1222,8 +1289,7 @@ int prepare_mappings(struct pstree_item *t)
 		if (ret < 0)
 			pr_perror("Unable to unmap %p(%lx)", addr, tail);
 		rsti(t)->premmapped_len = old_premmapped_len;
-		pr_info("Shrunk premap area to %p(%lx)\n",
-				rsti(t)->premmapped_addr, rsti(t)->premmapped_len);
+		pr_info("Shrunk premap area to %p(%lx)\n", rsti(t)->premmapped_addr, rsti(t)->premmapped_len);
 	}
 
 out:
@@ -1276,9 +1342,8 @@ int open_vmas(struct pstree_item *t)
 		if (!vma_area_is(vma, VMA_AREA_REGULAR) || !vma->vm_open)
 			continue;
 
-		pr_info("Opening %#016"PRIx64"-%#016"PRIx64" %#016"PRIx64" (%x) vma\n",
-				vma->e->start, vma->e->end,
-				vma->e->pgoff, vma->e->status);
+		pr_info("Opening %#016" PRIx64 "-%#016" PRIx64 " %#016" PRIx64 " (%x) vma\n", vma->e->start,
+			vma->e->end, vma->e->pgoff, vma->e->status);
 
 		if (vma->vm_open(pid, vma)) {
 			pr_err("`- Can't open vma\n");
@@ -1290,8 +1355,7 @@ int open_vmas(struct pstree_item *t)
 		 * turn, puts the VMA_CLOSE bit itself. For all the rest we
 		 * need to put it by hands, so that the restorer closes the fd
 		 */
-		if (!(vma_area_is(vma, VMA_FILE_PRIVATE) ||
-					vma_area_is(vma, VMA_FILE_SHARED)))
+		if (!(vma_area_is(vma, VMA_FILE_PRIVATE) || vma_area_is(vma, VMA_FILE_SHARED)))
 			vma->e->status |= VMA_CLOSE;
 	}
 
@@ -1304,12 +1368,25 @@ static int prepare_vma_ios(struct pstree_item *t, struct task_restore_args *ta)
 {
 	struct cr_img *pages;
 
+	/*
+	 * We optimize the case when rsti(t)->vma_io is empty.
+	 *
+	 * This is useful when using the image streamer, where all VMAs are
+	 * premapped (pr->pieok is false). This avoids re-opening the
+	 * CR_FD_PAGES file, which may only be readable only once.
+	 */
+	if (list_empty(&rsti(t)->vma_io)) {
+		ta->vma_ios = NULL;
+		ta->vma_ios_n = 0;
+		ta->vma_ios_fd = -1;
+		return 0;
+	}
+
 	/*
 	 * If auto-dedup is on we need RDWR mode to be able to punch holes in
 	 * the input files (in restorer.c)
 	 */
-	pages = open_image(CR_FD_PAGES, opts.auto_dedup ? O_RDWR : O_RSTR,
-				rsti(t)->pages_img_id);
+	pages = open_image(CR_FD_PAGES, opts.auto_dedup ? O_RDWR : O_RSTR, rsti(t)->pages_img_id);
 	if (!pages)
 		return -1;
 
