@@ -1,5 +1,4 @@
 #!/usr/bin/env python
-# vim: noet ts=8 sw=8 sts=8
 from __future__ import absolute_import, division, print_function, unicode_literals
 
 import argparse
@@ -22,6 +21,7 @@ import subprocess
 import sys
 import tempfile
 import time
+import socket
 from builtins import (input, int, open, range, str, zip)
 
 import pycriu as crpc
@@ -30,6 +30,9 @@ import yaml
 
 os.chdir(os.path.dirname(os.path.abspath(__file__)))
 
+# File to store content of streamed images
+STREAMED_IMG_FILE_NAME = "img.criu"
+
 prev_line = None
 
 
@@ -888,6 +891,8 @@ class criu_rpc_process:
 
 
 class criu_rpc:
+    pidfd_store_socket = None
+
     @staticmethod
     def __set_opts(criu, args, ctx):
         while len(args) != 0:
@@ -941,6 +946,10 @@ class criu_rpc:
                 fd, key = key.split(":", 1)
                 inhfd.fd = int(fd[3:-1])
                 inhfd.key = key
+            elif "--pidfd-store" == arg:
+                if criu_rpc.pidfd_store_socket is None:
+                    criu_rpc.pidfd_store_socket = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
+                criu.opts.pidfd_store_sk = criu_rpc.pidfd_store_socket.fileno()
             else:
                 raise test_fail_exc('RPC for %s(%s) required' % (arg, args.pop(0)))
 
@@ -1024,6 +1033,7 @@ class criu:
         self.__mdedup = bool(opts['noauto_dedup'])
         self.__user = bool(opts['user'])
         self.__leave_stopped = bool(opts['stop'])
+        self.__stream = bool(opts['stream'])
         self.__criu = (opts['rpc'] and criu_rpc or criu_cli)
         self.__show_stats = bool(opts['show_stats'])
         self.__lazy_pages_p = None
@@ -1039,19 +1049,19 @@ class criu:
             ret = self.__dump_process.wait()
         if self.__lazy_pages_p:
             ret = self.__lazy_pages_p.wait()
-            grep_errors(os.path.join(self.__ddir(), "lazy-pages.log"))
+            grep_errors(os.path.join(self.__ddir(), "lazy-pages.log"), err=ret)
             self.__lazy_pages_p = None
             if ret:
                 raise test_fail_exc("criu lazy-pages exited with %s" % ret)
         if self.__page_server_p:
             ret = self.__page_server_p.wait()
-            grep_errors(os.path.join(self.__ddir(), "page-server.log"))
+            grep_errors(os.path.join(self.__ddir(), "page-server.log"), err=ret)
             self.__page_server_p = None
             if ret:
                 raise test_fail_exc("criu page-server exited with %s" % ret)
         if self.__dump_process:
             ret = self.__dump_process.wait()
-            grep_errors(os.path.join(self.__ddir(), "dump.log"))
+            grep_errors(os.path.join(self.__ddir(), "dump.log"), err=ret)
             self.__dump_process = None
             if ret:
                 raise test_fail_exc("criu dump exited with %s" % ret)
@@ -1208,11 +1218,19 @@ class criu:
             stats_written = int(stent['shpages_written']) + int(
                 stent['pages_written'])
 
+        if self.__stream:
+            p = self.spawn_criu_image_streamer("extract")
+            p.wait()
+
         real_written = 0
         for f in os.listdir(self.__ddir()):
             if f.startswith('pages-'):
                 real_written += os.path.getsize(os.path.join(self.__ddir(), f))
 
+        if self.__stream:
+            # make sure the extracted image is not usable.
+            os.unlink(os.path.join(self.__ddir(), "inventory.img"))
+
         r_pages = real_written / mmap.PAGESIZE
         r_off = real_written % mmap.PAGESIZE
         if (stats_written != r_pages) or (r_off != 0):
@@ -1220,6 +1238,57 @@ class criu:
                   (stats_written, r_pages, r_off))
             raise test_fail_exc("page counts mismatch")
 
+    # action can be "capture", "extract", or "serve"
+    def spawn_criu_image_streamer(self, action):
+        print("Run criu-image-streamer in {} mode".format(action))
+
+        progress_r, progress_w = os.pipe()
+        # We fcntl() on both file descriptors due to some potential differences
+        # with python2 and python3.
+        fcntl.fcntl(progress_r, fcntl.F_SETFD, fcntl.FD_CLOEXEC)
+        fcntl.fcntl(progress_w, fcntl.F_SETFD, 0)
+
+        # We use cat because the streamer requires to work with pipes.
+        if action == 'capture':
+            cmd = ["criu-image-streamer",
+                   "--images-dir '{images_dir}'",
+                   "--progress-fd {progress_fd}",
+                   action,
+                   "| cat > {img_file}"]
+        else:
+            cmd = ["cat {img_file} |",
+                   "criu-image-streamer",
+                   "--images-dir '{images_dir}'",
+                   "--progress-fd {progress_fd}",
+                   action]
+
+        # * As we are using a shell pipe command, we want to use pipefail.
+        # Otherwise, failures stay unnoticed. For this, we use bash as sh
+        # doesn't support that feature.
+        # * We use close_fds=False because we want the child to inherit the progress pipe
+        p = subprocess.Popen(["bash", "-c", "set -o pipefail; " + " ".join(cmd).format(
+            progress_fd=progress_w,
+            images_dir=self.__ddir(),
+            img_file=os.path.join(self.__ddir(), STREAMED_IMG_FILE_NAME)
+        )], close_fds=False)
+
+        os.close(progress_w)
+        progress = os.fdopen(progress_r, "r")
+
+        if action == 'serve' or action == 'extract':
+            # Consume image statistics
+            progress.readline()
+
+        if action == 'capture' or action == 'serve':
+            # The streamer socket is ready for consumption once we receive the
+            # socket-init message.
+            if progress.readline().strip() != "socket-init":
+                p.kill()
+                raise test_fail_exc(
+                    "criu-image-streamer is not starting (exit_code=%d)" % p.wait())
+
+        return p
+
     def dump(self, action, opts=[]):
         self.__iter += 1
         os.mkdir(self.__ddir())
@@ -1249,6 +1318,10 @@ class criu:
 
         a_opts += self.__test.getdopts()
 
+        if self.__stream:
+            streamer_p = self.spawn_criu_image_streamer("capture")
+            a_opts += ["--stream"]
+
         if self.__dedup:
             a_opts += ["--auto-dedup"]
 
@@ -1273,6 +1346,11 @@ class criu:
         self.__dump_process = self.__criu_act(action,
                                               opts=a_opts + opts,
                                               nowait=nowait)
+        if self.__stream:
+            ret = streamer_p.wait()
+            if ret:
+                raise test_fail_exc("criu-image-streamer exited with %d" % ret)
+
         if self.__mdedup and self.__iter > 1:
             self.__criu_act("dedup", opts=[])
 
@@ -1285,7 +1363,7 @@ class criu:
 
         if self.__page_server_p:
             ret = self.__page_server_p.wait()
-            grep_errors(os.path.join(self.__ddir(), "page-server.log"))
+            grep_errors(os.path.join(self.__ddir(), "page-server.log"), err=ret)
             self.__page_server_p = None
             if ret:
                 raise test_fail_exc("criu page-server exited with %d" % ret)
@@ -1303,6 +1381,10 @@ class criu:
             r_opts += ['--empty-ns', 'net']
             r_opts += ['--action-script', os.getcwd() + '/empty-netns-prep.sh']
 
+        if self.__stream:
+            streamer_p = self.spawn_criu_image_streamer("serve")
+            r_opts += ["--stream"]
+
         if self.__dedup:
             r_opts += ["--auto-dedup"]
 
@@ -1336,6 +1418,11 @@ class criu:
             r_opts += ['--leave-stopped']
 
         self.__criu_act("restore", opts=r_opts + ["--restore-detached"])
+        if self.__stream:
+            ret = streamer_p.wait()
+            if ret:
+                raise test_fail_exc("criu-image-streamer exited with %d" % ret)
+
         self.show_stats("restore")
 
         if self.__leave_stopped:
@@ -1344,6 +1431,13 @@ class criu:
 
     @staticmethod
     def check(feature):
+        if feature == 'stream':
+            try:
+                p = subprocess.Popen(["criu-image-streamer", "--version"])
+                return p.wait() == 0
+            except Exception:
+                return False
+
         return criu_cli.run(
             "check", ["--no-default-config", "-v0", "--feature", feature],
             opts['criu_bin']) == 0
@@ -1400,7 +1494,7 @@ def init_sbs():
 
 def sbs(what):
     if do_sbs:
-        input("Pause at %s. Press Enter to continue." % what)
+        input("Pause %s. Press Enter to continue." % what)
 
 
 #
@@ -1422,14 +1516,16 @@ def cr(cr_api, test, opts):
         pres = iter_parm(opts['pre'], 0)
         for p in pres[0]:
             if opts['snaps']:
+                sbs('before snap %d' % p)
                 cr_api.dump("dump", opts=["--leave-running", "--track-mem"])
             else:
+                sbs('before pre-dump %d' % p)
                 cr_api.dump("pre-dump")
                 try_run_hook(test, ["--post-pre-dump"])
                 test.pre_dump_notify()
             time.sleep(pres[1])
 
-        sbs('pre-dump')
+        sbs('before dump')
 
         os.environ["ZDTM_TEST_PID"] = str(test.getpid())
         if opts['norst']:
@@ -1442,13 +1538,13 @@ def cr(cr_api, test, opts):
                 test.gone()
             else:
                 test.unlink_pidfile()
-            sbs('pre-restore')
+            sbs('before restore')
             try_run_hook(test, ["--pre-restore"])
             cr_api.restore()
             os.environ["ZDTM_TEST_PID"] = str(test.getpid())
             os.environ["ZDTM_IMG_DIR"] = cr_api.logs()
             try_run_hook(test, ["--post-restore"])
-            sbs('post-restore')
+            sbs('after restore')
 
         time.sleep(iters[1])
 
@@ -1579,6 +1675,37 @@ class noop_freezer:
         return []
 
 
+class cg_freezer2:
+    def __init__(self, path, state):
+        self.__path = '/sys/fs/cgroup/' + path
+        self.__state = state
+        self.kernel = True
+
+    def attach(self):
+        if not os.access(self.__path, os.F_OK):
+            os.makedirs(self.__path)
+        with open(self.__path + '/cgroup.procs', 'w') as f:
+            f.write('0')
+
+    def __set_state(self, state):
+        with open(self.__path + '/cgroup.freeze', 'w') as f:
+            f.write(state)
+
+    def freeze(self):
+        if self.__state.startswith('f'):
+            self.__set_state('1')
+
+    def thaw(self):
+        if self.__state.startswith('f'):
+            self.__set_state('0')
+
+    def getdopts(self):
+        return ['--freeze-cgroup', self.__path, '--manage-cgroups']
+
+    def getropts(self):
+        return ['--manage-cgroups']
+
+
 class cg_freezer:
     def __init__(self, path, state):
         self.__path = '/sys/fs/cgroup/freezer/' + path
@@ -1615,7 +1742,11 @@ def get_freezer(desc):
         return noop_freezer()
 
     fd = desc.split(':')
-    fr = cg_freezer(path=fd[0], state=fd[1])
+
+    if os.access("/sys/fs/cgroup/user.slice/cgroup.procs", os .F_OK):
+        fr = cg_freezer2(path=fd[0], state=fd[1])
+    else:
+        fr = cg_freezer(path=fd[0], state=fd[1])
     return fr
 
 
@@ -1809,7 +1940,7 @@ class Launcher:
             self.__taint = taintfd.read()
         if int(self.__taint, 0) != 0:
             print("The kernel is tainted: %r" % self.__taint)
-            if not opts["ignore_taint"]:
+            if not opts["ignore_taint"] and os.getenv("ZDTM_IGNORE_TAINT") != '1':
                 raise Exception("The kernel is tainted: %r" % self.__taint)
 
     def __show_progress(self, msg):
@@ -1852,7 +1983,7 @@ class Launcher:
               'stop', 'empty_ns', 'fault', 'keep_img', 'report', 'snaps',
               'sat', 'script', 'rpc', 'lazy_pages', 'join_ns', 'dedup', 'sbs',
               'freezecg', 'user', 'dry_run', 'noauto_dedup',
-              'remote_lazy_pages', 'show_stats', 'lazy_migrate',
+              'remote_lazy_pages', 'show_stats', 'lazy_migrate', 'stream',
               'tls', 'criu_bin', 'crit_bin', 'pre_dump_mode')
         arg = repr((name, desc, flavor, {d: self.__opts[d] for d in nd}))
 
@@ -2042,16 +2173,16 @@ def print_error(line):
     return False
 
 
-def grep_errors(fname):
+def grep_errors(fname, err=False):
     first = True
     print_next = False
     before = []
     with open(fname, errors='replace') as fd:
-        for l in fd:
-            before.append(l)
+        for line in fd:
+            before.append(line)
             if len(before) > 5:
                 before.pop(0)
-            if "Error" in l or "Warn" in l:
+            if "Error" in line or "Warn" in line:
                 if first:
                     print_fname(fname, 'log')
                     print_sep("grep Error", "-", 60)
@@ -2061,8 +2192,19 @@ def grep_errors(fname):
                 before = []
             else:
                 if print_next:
-                    print_next = print_error(l)
+                    print_next = print_error(line)
                     before = []
+
+    # If process failed but there are no errors in log,
+    # let's just print the log tail, probably it would
+    # be helpful.
+    if err and first:
+        print_fname(fname, 'log')
+        print_sep("grep Error (no)", "-", 60)
+        first = False
+        for i in before:
+            print_next = print_error(i)
+
     if not first:
         print_sep("ERROR OVER", "-", 60)
 
@@ -2138,6 +2280,15 @@ def run_tests(opts):
                 "[WARNING] Non-cooperative UFFD is missing, some tests might spuriously fail"
             )
 
+    if opts['stream']:
+        streamer_dir = os.path.realpath(opts['criu_image_streamer_dir'])
+        os.environ['PATH'] = "{}:{}".format(streamer_dir, os.environ['PATH'])
+        if not criu.check('stream'):
+            raise RuntimeError((
+                "Streaming tests need the criu-image-streamer binary to be accessible in the {} directory. " +
+                "Specify --criu-image-streamer-dir or modify PATH to provide an alternate location")
+                .format(streamer_dir))
+
     launcher = Launcher(opts, len(torun))
     try:
         for t in torun:
@@ -2460,8 +2611,8 @@ rp.add_argument("--rpc",
 rp.add_argument("--page-server",
                 help="Use page server dump",
                 action='store_true')
-rp.add_argument("--remote",
-                help="Use remote option for diskless C/R",
+rp.add_argument("--stream",
+                help="Use criu-image-streamer",
                 action='store_true')
 rp.add_argument("-p", "--parallel", help="Run test in parallel")
 rp.add_argument("--dry-run",
@@ -2500,6 +2651,9 @@ rp.add_argument("--criu-bin",
 rp.add_argument("--crit-bin",
                 help="Path to crit binary",
                 default='../crit/crit')
+rp.add_argument("--criu-image-streamer-dir",
+                help="Directory where the criu-image-streamer binary is located",
+                default="../../criu-image-streamer")
 rp.add_argument("--pre-dump-mode",
                 help="Use splice or read mode of pre-dumping",
                 choices=['splice', 'read'],
